﻿using PalmSens;
using PalmSens.Comm;
using PalmSens.Core.Simplified.Data;
using PalmSens.Devices;
using PalmSens.Techniques;
using System;
using System.Collections.Generic;
using System.Threading.Tasks;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;

// The Blank Page item template is documented at http://go.microsoft.com/fwlink/?LinkId=402352&clcid=0x409

namespace PSSDKBiPotAuxExample
{
    /// <summary>
    /// An empty page that can be used on its own or navigated to within a Frame.
    /// </summary>
    public sealed partial class MainPage : Page
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="MainPage"/> class.
        /// </summary>
        public MainPage()
        {
            this.InitializeComponent();
        }

        /// <summary>
        /// Handles the Loaded event of the Page control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="RoutedEventArgs"/> instance containing the event data.</param>
        private async void Page_Loaded(object sender, RoutedEventArgs e)
        {
            await DiscoverConnectedDevices();
            InitResultsListBox();
            InitCAMethod();
            cmbBiPotMode.SelectedIndex = 0;
            cmbExtra.SelectedIndex = 1;
        }

        /// <summary>
        /// The channels bound to the cmbChannels control
        /// </summary>
        List<string> ExtraValueType = new List<string>() { "BiPot", "Aux" };

        List<string> BiPotModes = new List<string>() { "Constant", "Offset" };

        /// <summary>
        /// The instance of method class containing the Chronoamperometry parameters with BiPot enabled
        /// </summary>
        private AmperometricDetection _methodCABiPot;

        /// <summary>
        /// The instance of method class containing the Chronoamperometry parameters with Aux Recording enabled
        /// </summary>
        private AmperometricDetection _methodCAAux;

        /// <summary>
        /// The connected PalmSens & EmStat devices
        /// </summary>
        private Device[] _connectedDevices = new Device[0];

        /// <summary>
        /// The active SimpleMeasurement
        /// </summary>
        private SimpleMeasurement _activeMeasurement = null;

        /// <summary>
        /// The extra value curve
        /// </summary>
        private SimpleCurve _extraValueCurve = null;

        /// <summary>
        /// Initializes the listbox control.
        /// </summary>
        private void InitResultsListBox()
        {
            lbResults.Items.Clear();
            lbResults.Items.Add("ID     Time (s)   Current (µA)    Extra Value");
        }

        /// <summary>
        /// Initializes the CA method.
        /// </summary>
        private void InitCAMethod()
        {
            //BiPot Method
            _methodCABiPot = new AmperometricDetection(); //Create a new chronoamperometry method with the default settings
            _methodCABiPot.Potential = 1f; //Sets the potential at 1V
            _methodCABiPot.IntervalTime = 0.1f; //Sets the sampling interval time at 0.1s
            _methodCABiPot.RunTime = 3f; //Sets the measurement runtime at 3s

            _methodCABiPot.EquilibrationTime = 1f; //Equilabrates the cell at the defined potential for 1 second before starting the measurement
            _methodCABiPot.Ranging.StartCurrentRange = new CurrentRange(CurrentRanges.cr1uA); //Starts equilabration in the 1µA current range
            _methodCABiPot.Ranging.MinimumCurrentRange = new CurrentRange(CurrentRanges.cr10nA); //Min current range 10nA
            _methodCABiPot.Ranging.MaximumCurrentRange = new CurrentRange(CurrentRanges.cr1mA); //Max current range 1mA

            //BiPot settings
            _methodCABiPot.ExtraValueMsk = ExtraValueMask.BipotWE;
            //The BiPot mode and potential are updated with the settings specified in the Form in btnMeasureBiPot_Click
            _methodCABiPot.BipotModePS = Method.EnumPalmSensBipotMode.offset;
            _methodCABiPot.BiPotPotential = 0;
            _methodCABiPot.BiPotCR = new CurrentRange(CurrentRanges.cr1uA); //Set BiPot to measure in the 1µA current range

            //Set BiPot to measure in the 1µA current range//Init EmStat Pico settings, these do not affect other devices (alternatively use the method _methodCABiPot.DeterminePGStatMode((EmStatPicoCapabilities)_psCommSimpleAndroid.Capabilities) to determine the recommended settings for the EmStat Pico)
            _methodCABiPot.PGStatMode = MethodScript.PGStatModes.LowSpeed; //BiPot measurements require the device to be set to lowspeed mode
            _methodCABiPot.SelectedPotentiostatChannel = PotentionstatChannels.Ch0; //In BiPot measurements the main we, the re and ce are controlled by channel 0 and the BiPot we by channel 1

            //Aux Method
            _methodCAAux = new AmperometricDetection(); //Create a new chronoamperometry method with the default settings
            _methodCAAux.Potential = 1f; //Sets the potential at 1V
            _methodCAAux.IntervalTime = 0.1f; //Sets the sampling interval time at 0.1s
            _methodCAAux.RunTime = 3f; //Sets the measurement runtime at 3s

            _methodCAAux.EquilibrationTime = 1f; //Equilabrates the cell at the defined potential for 1 second before starting the measurement
            _methodCAAux.Ranging.StartCurrentRange = new CurrentRange(CurrentRanges.cr1uA); //Starts equilabration in the 1µA current range
            _methodCAAux.Ranging.MinimumCurrentRange = new CurrentRange(CurrentRanges.cr10nA); //Min current range 10nA
            _methodCAAux.Ranging.MaximumCurrentRange = new CurrentRange(CurrentRanges.cr1mA); //Max current range 1mA

            //Aux settings
            _methodCAAux.ExtraValueMsk = ExtraValueMask.AuxInput;
        }

        /// <summary>
        /// Discovers the connected PalmSens & EmStat devices and adds them to the combobox control.
        /// </summary>
        private async Task DiscoverConnectedDevices()
        {
            btnRefresh.IsEnabled = false;
            cmbDevices.Items.Clear();
            _connectedDevices = await psCommSimpleUWP.GetConnectedDevices(); //Discover connected devices

            foreach (Device d in _connectedDevices)
                cmbDevices.Items.Add(d.ToString()); //Add connected devices to control

            int nDevices = cmbDevices.Items.Count;
            cmbDevices.SelectedIndex = nDevices > 0 ? 0 : -1;
            lbConsole.Items.Add($"Found {nDevices} device(s).");

            btnConnect.IsEnabled = nDevices > 0;
            btnRefresh.IsEnabled = true;
        }

        /// <summary>
        /// Handles the Click event of the btnRefresh control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="RoutedEventArgs"/> instance containing the event data.</param>
        private async void btnRefresh_Click(object sender, RoutedEventArgs e)
        {
            await DiscoverConnectedDevices();
        }

        /// <summary>
        /// Handles the Click event of the btnConnect control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="RoutedEventArgs"/> instance containing the event data.</param>
        private async void btnConnect_Click(object sender, RoutedEventArgs e)
        {
            btnConnect.IsEnabled = false;
            if (!psCommSimpleUWP.Connected) //Determine whether a device is currently connected
            {
                if (cmbDevices.SelectedIndex == -1)
                    return;

                try
                {
                    //Connect to the device selected in the devices combobox control
                    await psCommSimpleUWP.Connect(_connectedDevices[cmbDevices.SelectedIndex]);
                    lbConsole.Items.Add($"Connected to {psCommSimpleUWP.ConnectedDevice.ToString()}");
                }
                catch (Exception ex)
                {
                    lbConsole.Items.Add(ex.Message);
                }
            }
            else
            {
                await psCommSimpleUWP.Disconnect(); //Disconnect from the connected device
                lbConsole.Items.Add("Disconnected");
            }

            //Update UI based on connection status
            cmbDevices.IsEnabled = !psCommSimpleUWP.Connected;
            btnRefresh.IsEnabled = !psCommSimpleUWP.Connected;
            btnConnect.Content = psCommSimpleUWP.Connected ? "Disconnect" : "Connect";
            btnMeasureAux.IsEnabled = psCommSimpleUWP.Connected;

            //Check device capabilities to see if BiPot is supported
            btnMeasureBiPot.IsEnabled = psCommSimpleUWP.Connected && psCommSimpleUWP.Capabilities.BiPotPresent;
            cmbExtra.IsEnabled = psCommSimpleUWP.Connected && psCommSimpleUWP.Capabilities.BiPotPresent;
            cmbExtra.SelectedIndex = (psCommSimpleUWP.Connected && psCommSimpleUWP.Capabilities.BiPotPresent) ? 0 : 1;

            btnConnect.IsEnabled = true;
        }

        /// <summary>
        /// Handles the Click event of the btnMeasureBiPot control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="RoutedEventArgs"/> instance containing the event data.</param>
        private async void btnMeasureBiPot_Click(object sender, RoutedEventArgs e)
        {
            btnMeasureBiPot.IsEnabled = btnMeasureAux.IsEnabled = false;
            if (psCommSimpleUWP.DeviceState == PalmSens.Comm.CommManager.DeviceState.Idle) //Determine whether the device is currently idle or measuring
            {
                try
                {
                    //Parse the BiPot potential specified in the textbox
                    float potentialBiPot;
                    if (!float.TryParse(tbBiPotPotential.Text, out potentialBiPot))
                    {
                        lbConsole.Items.Add("Could not parse BiPot potential.");
                        return;
                    }
                    _methodCABiPot.BiPotPotential = potentialBiPot; //Set the BiPot potential specified in the form
                    //Set the BiPot mode specified in the form
                    _methodCABiPot.BipotModePS = cmbBiPotMode.SelectedIndex == 0 ? Method.EnumPalmSensBipotMode.constant : Method.EnumPalmSensBipotMode.offset;

                    InitResultsListBox(); //Reset the data list box
                    //Start BiPot measurement defined in the method
                    _activeMeasurement = await psCommSimpleUWP.Measure(_methodCABiPot);

                    //Create a new SimpleCurve with ExtraValue(BiPot/Aux) readings (Set to silent to prevent the new data event from being raised by multiple curves)
                    _extraValueCurve = (_activeMeasurement.NewSimpleCurve(PalmSens.Data.DataArrayType.Time, PalmSens.Data.DataArrayType.BipotCurrent, "", true))[0];
                }
                catch (Exception ex)
                {
                    lbConsole.Items.Add(ex.Message);
                }
            }
            else
            {
                try
                {
                    await psCommSimpleUWP.AbortMeasurement(); //Abort the active measurement
                }
                catch (Exception ex)
                {
                    lbConsole.Items.Add(ex.Message);
                }
            }
            btnMeasureBiPot.IsEnabled = btnMeasureAux.IsEnabled = true;
        }

        /// <summary>
        /// Handles the Click event of the btnMeasureAux control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="RoutedEventArgs"/> instance containing the event data.</param>
        private async void btnMeasureAux_Click(object sender, RoutedEventArgs e)
        {
            btnMeasureBiPot.IsEnabled = btnMeasureAux.IsEnabled = true;
            if (psCommSimpleUWP.DeviceState == PalmSens.Comm.CommManager.DeviceState.Idle) //Determine whether the device is currently idle or measuring
            {
                try
                {
                    InitResultsListBox(); //Reset the data list box
                    //Start measurement with auxillary readings enabled defined in the method
                    _activeMeasurement = await psCommSimpleUWP.Measure(_methodCAAux);

                    //Create a new SimpleCurve with ExtraValue(BiPot/Aux) readings (Set to silent to prevent the new data event from being raised by multiple curves)
                    _extraValueCurve = (_activeMeasurement.NewSimpleCurve(PalmSens.Data.DataArrayType.Time, PalmSens.Data.DataArrayType.AuxInput, "", true))[0];
                }
                catch (Exception ex)
                {
                    lbConsole.Items.Add(ex.Message);
                }
            }
            else
            {
                try
                {
                    await psCommSimpleUWP.AbortMeasurement(); //Abort the active measurement
                }
                catch (Exception ex)
                {
                    lbConsole.Items.Add(ex.Message);
                }
            }
            btnMeasureBiPot.IsEnabled = btnMeasureAux.IsEnabled = true;
        }

        /// <summary>
        /// Raised when the measurement is ended
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void psCommSimpleUWP_MeasurementEnded(object sender, EventArgs e)
        {
            lbConsole.Items.Add("Measurement ended.");
        }

        /// <summary>
        /// Raised when the measurement is started
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void psCommSimpleUWP_MeasurementStarted(object sender, EventArgs e)
        {
            lbConsole.Items.Add("Measurement started.");
        }

        /// <summary>
        /// Raised when device status package is received (the device does not send status packages while measuring)
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="PalmSens.Comm.StatusEventArgs"/> instance containing the event data.</param>
        private void psCommSimpleUWP_ReceiveStatus(object sender, PalmSens.Comm.StatusEventArgs e)
        {
            Status status = e.GetStatus(); //Get the PalmSens.Comm.Status instance from the event data
            double potential = status.PotentialReading.Value; //Get the potential
            double currentInRange = status.CurrentReading.ValueInRange; //Get the current expressed inthe active current range
            PalmSens.Comm.ReadingStatus currentStatus = status.CurrentReading.ReadingStatus; //Get the status of the current reading
            CurrentRange cr = status.CurrentReading.CurrentRange; //Get the active current range

            tbPotential.Text = potential.ToString("F3");
            tbCurrent.Text = currentInRange.ToString("F3");
            lblCurrentRange.Text = $"* {cr.ToString()}";

            if (cmbExtra.SelectedIndex == 0)
            {
                tbExtraValue.Text = status.GetCorrectedBiPotCurrent().ToString("F3");
                lblExtraValue.Text = lblCurrentRange.Text;
            }
            else
            {
                tbExtraValue.Text = status.GetAuxInputAsVoltage().ToString("F3");
                lblExtraValue.Text = "V";
            }
        }

        /// <summary>
        /// Raised when the connected device's status changes
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="CurrentState">State of the current.</param>
        private void psCommSimpleUWP_StateChanged(object sender, PalmSens.Comm.CommManager.DeviceState CurrentState)
        {
            tbStatus.Text = CurrentState.ToString(); //Updates the device state indicator textbox
            btnConnect.IsEnabled = CurrentState == PalmSens.Comm.CommManager.DeviceState.Idle;
            btnMeasureBiPot.Content = CurrentState == CommManager.DeviceState.Idle ? "Measure" : "Abort";
            btnMeasureAux.Content = CurrentState == CommManager.DeviceState.Idle ? "Measure" : "Abort";
        }

        /// <summary>
        /// Raised when a Simple Curve in the active SimpleMeasurement starts receiving data
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="activeSimpleCurve">The active simple curve.</param>
        private void psCommSimpleUWP_SimpleCurveStartReceivingData(object sender, SimpleCurve activeSimpleCurve)
        {
            if (activeSimpleCurve.YAxisDataType != PalmSens.Data.DataArrayType.Current) //Ignore other automatically generated curves
                return;

            //Subscribe to the curve's events to receive updates when new data is available and when it iss finished receiving data
            activeSimpleCurve.NewDataAdded += ActiveCurve_NewDataAdded;
            activeSimpleCurve.CurveFinished += ActiveCurve_CurveFinished;

            lbConsole.Items.Add($"Curve is receiving new data...");
        }

        /// <summary>
        /// Raised when new data points are added to the active SimpleCurve
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="PalmSens.Data.ArrayDataAddedEventArgs"/> instance containing the event data.</param>
        private void ActiveCurve_NewDataAdded(object sender, PalmSens.Data.ArrayDataAddedEventArgs e)
        {
            SimpleCurve activeCurve = sender as SimpleCurve;
            int startIndex = e.StartIndex; //The index of the first new data point added to the curve
            int count = e.Count; //The number of new data points added to the curve

            for (int i = startIndex; i < startIndex + count; i++)
            {
                double xValue = activeCurve.XAxisValue(i); //Get the value on Curve's X-Axis (potential) at the specified index
                double yValue = activeCurve.YAxisValue(i); //Get the value on Curve's Y-Axis (current) at the specified index

                //Get the BiPot/Aux reading
                double extraValue = _extraValueCurve.YAxisValue(i); //BiPot
                string extraValueUnit = _activeMeasurement.Measurement.Method.ExtraValueMsk == ExtraValueMask.BipotWE ? "µA" : "V";

                lbResults.Items.Add($"{i+1}     {xValue.ToString("F2")}     {yValue.ToString("E3")}     {extraValue.ToString("E3")} {extraValueUnit}");
            }
        }

        /// <summary>
        /// Raised when a SimpleCurve stops receiving new data points
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void ActiveCurve_CurveFinished(object sender, EventArgs e)
        {
            SimpleCurve activeCurve = sender as SimpleCurve;

            int nDataPointsReceived = activeCurve.NDataPoints;
            lbConsole.Items.Add($"Channel {activeCurve.MuxChannel + 1}: {nDataPointsReceived} data point(s) received.");

            //Unsubscribe from the curves events to avoid memory leaks
            activeCurve.NewDataAdded -= ActiveCurve_NewDataAdded;
            activeCurve.CurveFinished -= ActiveCurve_CurveFinished;

            lbConsole.Items.Add($"Curve Finished");
        }
    }
}
